/*--------------------------------------------------------------------------
 * Copyright (C) 1992-1999 Dallas Semiconductor Corporation. 
 * All rights Reserved. Printed in U.S.A.
 * This software is protected by copyright laws of
 * the United States and of foreign countries.
 * This material may also be protected by patent laws of the United States 
 * and of foreign countries.
 * This software is furnished under a license agreement and/or a
 * nondisclosure agreement and may only be used or copied in accordance
 * with the terms of those agreements.
 * The mere transfer of this software does not imply any licenses
 * of trade secrets, proprietary technology, copyrights, patents,
 * trademarks, maskwork rights, or any other form of intellectual
 * property whatsoever. Dallas Semiconductor retains all ownership rights.
 *--------------------------------------------------------------------------
 *
 *  iBWeath.C - Test program for 1-Wire weather station.
 *              It uses the default microLan
 *              port with a standard weather station (only) connected.
 *              It is not compatible with the extra sensors such as a 
 *              rain gauge.
 *
 *  Compiler:  Microsoft Visual C 
 *  Version: 1.01
 *  History: 1.00 -> 1.01  Removed beeps, used different header file, 
 *                         use read default port API, use %ld when printing longs,
 *                         change references to DS2407 to DS2406.
 */

#define TMEXUTIL

#include "ibtmex.h"   // (1.01)   
#include <time.h>

/* constants */
#define TEMP         0
#define SWITCH       1
#define COUNT        2
#define TEMP_FAM     0x10
#define SWITCH_FAM   0x12
#define COUNT_FAM    0x1D

#define AOFF_BON     0x3F
#define AOFF_BOFF    0x7F

#define COUNTER_PAGE 15

char fam_zstr[][40] = 
     { { "Temperature (DS1820)" },
       { "Switch (DS2406)     " }, 
       { "Counter (DS2423)    " },
       { "" } };

/* Local Function Prototypes */
void Exit_Prog(char *, short);
short FindWeatherStation(void);
short FindFirstFamily(short family);
short ReadWeatherStation(void);
uchar dowcrc(uchar);
short ReadTempScratch(uchar *);
short ReadTemperature(float far *);
void main(short argc, char **argv);
void output(short print_flag, char *msg);
ushort docrc16(ushort);
short ReadSwitchInfo(short);
short SetFlipFlop(short);

/* Global Variables to hold list information */
short PortType, PortNum;
uchar state_buffer[5120];
short WeatherROM[3][8];
long session_handle;                         
ushort CRC16;
uchar CRC8;                   
short Verbose=0;
FILE *fp;

/*----------------------------------------------------------------------
 *  This is the Main routine for iBWeath
 */
void main(short argc, char **argv)
{
   short printmsg=1,found,ClearActivity=0,filenum;
   short foundstation=0,didsetup=0,rslt,errorcnt=0;
   char tstr[80],temp_zstr[80]; 
         

   /* check arguments to see if request instruction with '?' or too many */
   if ((argc > 3) || ((argc > 1) && (argv[1][0] == '?' || argv[1][1] == '?')))
   {
       printf("\nusage: DiBWeath <output_filename> </Verbose>\n"
              "  - 1-Wire Weather station test on the default\n"
              "    MicroLAN port\n"
              "  - <output_filename> optional output filename\n"
              "  - </Verbose> optional verbose mode\n"
              "  - press any key to stop weather station test\n"
              "  - version 1.01\n");
       Exit_Prog("",BAD_ARGUMENT);
   }

   /* read the default PortNum and PortType */ 
   if (!TMReadDefaultPort(&PortNum,&PortType))
   {
       printf("ERROR, Could not read the default PortNum and PortType!\n");
       exit(1);
   }

   /* check arguments */
   Verbose = FALSE;
   filenum = 0;
   if (argc >= 2)
   {
      if (argv[1][0] != '/')
         filenum = 1;
      else if ((argv[1][1] == 'V') || (argv[1][1] == 'v'))
         Verbose = TRUE;

      if (argc == 3)
      {    
         if (argv[2][0] != '/')
            filenum = 2;
         else if ((argv[2][1] == 'V') || (argv[2][1] == 'v'))
            Verbose = TRUE;
      }   
   }

   /* open the output file */ 
   fp = NULL;
   if (filenum > 0)
   {
      fp = fopen(argv[filenum],"w+");
      if(fp == NULL)
         Exit_Prog("ERROR, Could not open output file!\n",0);
   }

   /* print a header */  
   output(TRUE,"\niBWeath Version 1.01\n"
         "Dallas Semiconductor Corporation\n"
         "Copyright (C) 1999\n\n");
   sprintf(temp_zstr,"Default Port number: %d    Port type: %d\n",PortNum,PortType);
   output(Verbose,temp_zstr);   
   Get_Version(tstr);
   sprintf(temp_zstr,"TMEX main driver: %s\n",tstr);
   output(Verbose,temp_zstr);   
   sprintf(temp_zstr,"TMEX adapter driver #%d:",PortType);
   output(Verbose,temp_zstr);   
   if (TMGetTypeVersion(PortType,tstr) < 0)
      Exit_Prog("No Hardware Driver for this type found!",NO_DRIVERS);
   sprintf(temp_zstr," %s\n\n",tstr);
   output(Verbose,temp_zstr);   
   output((short)!Verbose,"\nFormat (Time, Temperature(F), Direction ROM's, Wind speed(MPH))\n\n");

   /* loop to get a session handle, find weather station, read */
   found = FALSE;
   do
   {
      /* attempt to get a session */
      session_handle = TMExtendedStartSession(PortNum,PortType,NULL);
      if (session_handle > 0)  
      {
         /* reset the session prompt */
         printmsg = TRUE;
 
         /* setup the MicroLan */
         if (!didsetup)
         {
            rslt = TMSetup(session_handle);
            sprintf(temp_zstr,"Result of setup of the MicroLAN: %d\n\n",rslt);
            output(Verbose,temp_zstr);
            /* check for good setup */
            if (rslt == 1)
               didsetup = TRUE;
            else
               break;
         }

         /* find the Weather Station */
         if (didsetup)
         {
            /* check if station is found */
            if (!foundstation)
            {
               if (!FindWeatherStation())
               {
                  output(TRUE, "\nCould not find all components required for weather station.\n");
                  break;
               }
               else
                  foundstation = TRUE;
            }

            /* read the weather station */
            if (ReadWeatherStation())
               errorcnt = 0;   /* reset the error count */
            else
            {            
               errorcnt++;
               sprintf(temp_zstr,"Failed to read the weather station %d times in a row.\n",errorcnt);
               output(TRUE, temp_zstr);
               /* check for abort */
               if (errorcnt > 4)
                  break;
            }

         }   
      }
      else if (session_handle == 0)
      {                            
         /* check if need to print the waiting message */
         if (printmsg)
         {  
            printmsg = FALSE;
            output(Verbose,"\nWaiting to get access to the MicroLan\n\n");
         }
      }
      else if (session_handle < 0)   
      {
         output(TRUE,"No TMEX adapter driver for this type found!\n");
         break;
      }

      /* end the session if one is open */
      TMEndSession(session_handle);

      /* be friendly to other applications,sleep for 30ms */
      Sleep(30);
   }
   while (!kbhit());

   /* end the session if one is open */
   TMEndSession(session_handle);
}


/*----------------------------------------------------------------------
 * Find all of the components in a weather station:
 *  - DS1820
 *  - DS2423
 *  - DS2406
 */                                                          
short FindWeatherStation(void)
{    
   short i,j,desired_types[] = { TEMP_FAM, SWITCH_FAM, COUNT_FAM };
   char temp_zstr[256];

   /* loop to find each type */
   for (i = 0; i < 3; i++)
   {
      if (FindFirstFamily(desired_types[i]))
      {
         /* copy the rom number */
         WeatherROM[i][0] = 0;
         TMRom(session_handle,state_buffer,&WeatherROM[i][0]);

         sprintf(temp_zstr,"Weather Station %s 0x%02X found: ",fam_zstr[i],desired_types[i]);
         output(Verbose,temp_zstr);
         for (j = 7; j >= 0; j--)
         {
            sprintf(temp_zstr,"%02X",WeatherROM[i][j]);
            output(Verbose,temp_zstr);
         }
         output(Verbose,"\n");
      }
      else
      {
         sprintf(temp_zstr,"Could not find %s 0x%02X.\n",fam_zstr[i],desired_types[i]);
         output(Verbose,temp_zstr);
         return FALSE;
      }
   }

   /* found all three types */
   return TRUE;
}


/*----------------------------------------------------------------------
 * Read all of the weather station sensors 
 */                                                          
short ReadWeatherStation(void)
{
   float current_temp;
   double counts_per;
   char temp_zstr[256];
   short ROM[8],i,j,rslt;
   uchar rd_buf[50];
   unsigned long start_stamp,end_stamp,first_count=0, second_count=0;
   char timebuf[128]; 

   /* print the current time */
   output(Verbose,"\nReading Time: ");
   _tzset();    
   _strtime(timebuf);
   sprintf(temp_zstr,"%s",timebuf);  
   output(TRUE,temp_zstr);
   output(Verbose,"\n");
   output((short)!Verbose,",");

   /* read counter and take time stamp */
   TMRom(session_handle,state_buffer,&WeatherROM[COUNT][0]); 
   rslt = TMExtendedReadPage(session_handle,state_buffer,COUNTER_PAGE,
          rd_buf,0);
   start_stamp = GetTickCount();
   if (rslt == 40)
   {
      /* extract the counter */
      for (i = 35; i >= 32; i--)
      {
         first_count <<= 8;
         first_count |= rd_buf[i];
      }  

      sprintf(temp_zstr,"Start counter %ld, with timestamp %ld\n",first_count, start_stamp);
      output(Verbose,temp_zstr);
   }
   else
   {
      output(TRUE,"\nError reading DS2423 counter\n");
      sprintf(temp_zstr,"Counter present verification: %d\n",
              TMStrongAccess(session_handle,state_buffer));
      output(TRUE,temp_zstr);
      return FALSE;
   }

   /* read the temperature */
   TMRom(session_handle,state_buffer,&WeatherROM[TEMP][0]); 
   if (ReadTemperature(&current_temp))
   {
      sprintf(temp_zstr,"Temperature: (%5.1fC) ",current_temp);                                   
      output(Verbose,temp_zstr);
      sprintf(temp_zstr,"%5.1f",current_temp * 9 / 5 + 32);                                   
      output(TRUE,temp_zstr);
      output(Verbose,"F\n");
      output((short)!Verbose,", ");
   }
   else
   {
      output(TRUE,"Error reading the temperature\n");
      sprintf(temp_zstr,"Temperature present verification: %d\n",
              TMStrongAccess(session_handle,state_buffer));
      output(TRUE,temp_zstr);
      return FALSE;
   }

   /* read wind direction */
   /* turn channel B ON */
   TMRom(session_handle,state_buffer,&WeatherROM[SWITCH][0]); 
   if (SetFlipFlop(AOFF_BON))
      output(Verbose,"DS2406 switch B on\n");
   else
   {
      output(TRUE,"\nError setting DS2406 switch\n");
      sprintf(temp_zstr,"Switch present verification: %d\n",
              TMStrongAccess(session_handle,state_buffer));
      output(TRUE,temp_zstr);
      return FALSE;
   }
   /* let the present pulses go by */
   Sleep(10);
   /* recheck state in verbose mode */
   if (Verbose)
   {
      sprintf(temp_zstr,"Switch state 0x%02X, switch present %d\n",ReadSwitchInfo(TRUE),
             TMStrongAccess(session_handle,state_buffer));
      output(Verbose,temp_zstr);
   }      

   /* search for any DS2401's */
   TMFamilySearchSetup(session_handle,&state_buffer,0x01);
   output(Verbose,"Direction device(s): ");
   do
   {
      /* get first device in list with the specified family */
      if (TMNext(session_handle,state_buffer) == 1)
      {
         /* read the rom number */
         ROM[0] = 0;
         TMRom(session_handle,state_buffer,ROM);   
         
         /* check if correct type */
         if ((ROM[0] & 0x7F) == 0x01)
         {
            for (j = 7; j >= 0; j--)
            {
               sprintf(temp_zstr,"%02X",ROM[j]);
               output(TRUE,temp_zstr);
            }
            output(TRUE," ");
         }
         else
            break;
      }
      else
         break;
   }
   while ((ROM[0] & 0x7F) == 0x01);
   output((short)!Verbose,",");
   
   /* turn channel B OFF */
   TMRom(session_handle,state_buffer,&WeatherROM[SWITCH][0]); 
   if (SetFlipFlop(AOFF_BOFF))
      output(Verbose,"\nDS2406 switch B off\n");
   else
   {   
      output(TRUE,"\nError setting DS2406 switch\n");
      sprintf(temp_zstr,"Switch present verification: %d\n",TMStrongAccess(session_handle,state_buffer));
      output(TRUE,temp_zstr);
      return FALSE;
   }
   /* recheck state in verbose mode */
   if (Verbose)
   {
      sprintf(temp_zstr,"Switch state 0x%02X, switch present %d\n",ReadSwitchInfo(TRUE),
              TMStrongAccess(session_handle,state_buffer));
      output(Verbose,temp_zstr);
   }      

   /* read counter again and calculate rotations/second */
   TMRom(session_handle,state_buffer,&WeatherROM[COUNT][0]); 
   rslt = TMExtendedReadPage(session_handle,state_buffer,COUNTER_PAGE,
          rd_buf,0);
   end_stamp = GetTickCount();
   if (rslt == 40)
   {
      /* extract the counter */
      for (i = 35; i >= 32; i--)
      {
         second_count <<= 8;
         second_count |= rd_buf[i];
      }  

      /* caluculate the MPH from the counter information */
      sprintf(temp_zstr,"End counter %ld, with timestamp %ld giving %ld counts in %ld ms\n",
                        second_count, end_stamp, (second_count - first_count),
                        (end_stamp - start_stamp));
      output(Verbose,temp_zstr);

      counts_per = ((second_count - first_count) * 1000.0) / (end_stamp - start_stamp);
      sprintf(temp_zstr,"Counts/second: %5.1f, revolutions/second: %5.1f giving ",counts_per,counts_per / 2); 
      output(Verbose,temp_zstr);

      sprintf(temp_zstr,"%5.1f", counts_per / 2.0 * 2.453);
      output(TRUE,temp_zstr);
      output(Verbose,"MPH\n");
   }
   else
   {
      output(TRUE,"\nError reading DS2423 counter\n");
      sprintf(temp_zstr,"Counter present verification: %d\n",
              TMStrongAccess(session_handle,state_buffer));
      output(TRUE,temp_zstr);
      return FALSE;
   }

   output(TRUE,"\n");
   return TRUE;
}


/*----------------------------------------------------------------------
 * Find the first device with a particular family
 */
short FindFirstFamily(short family)
{
   short ROM[9];
   uchar rd_buf[50];
   
   /* set up to find the first device with family 'family */
   if (TMFamilySearchSetup(session_handle,&state_buffer,family) == 1)
   { 
      /* get first device in list with the specified family */
      if (TMNext(session_handle,state_buffer) == 1)
      {
         /* read the rom number */
         ROM[0] = 0;
         TMRom(session_handle,state_buffer,ROM);   
         
         /* check if correct type */
         if ((family & 0x7F) == (ROM[0] & 0x7F))
         {
            /* make sure if a DS2406 is found it is 2 channel */
            if (family == 0x12)
            {
               if (TMExtendedReadPage(session_handle,state_buffer,0,
                                  rd_buf,1) == 8)
               {
                  if (rd_buf[7] & 0x40)
                     return TRUE;
               }
            }
            else
               return TRUE;
         }
      }
   }
     
   /* failed to find device of that type */
   return FALSE;                              
}


/*----------------------------------------------------------------------
 * Get a list of the files on the 'num' device in the ROM list.
 */                                                          
short ReadTemperature(float far *current_temp)
{
   short flag,tsht;
   float tmp,cr,cpc;
   uchar rbuf[10];       
   unsigned long st;
                         
   /* access the device */
   if (TMAccess(session_handle,&state_buffer) != 1)
      return 0;
         
   /* send the Recall E2 command make sure Scratch is correct */
   TMTouchByte(session_handle,0xB8);

   /* send the start T command */
   if (TMAccess(session_handle,&state_buffer) != 1)
      return 0;
   
   /* prepare the strong pullup after next TMTouchByte */
   TMOneWireLevel(session_handle,LEVEL_SET,LEVEL_STRONG_PULLUP,PRIMED_BYTE);
         
   /* send the temperature conversion command */
   TMTouchByte(session_handle,0x44);
         
   /* sleep for a second */
   st = GetTickCount() + 500; 
   while (GetTickCount() < st)
   {
      Sleep(100);
      TMValidSession(session_handle);
   }

   /* disable the strong pull-up */
   TMOneWireLevel(session_handle,LEVEL_SET,LEVEL_NORMAL,PRIMED_NONE);
   
   /* verify conversion is complete */
   if (TMTouchBit(session_handle,0x01) != 0x01)
      return -100; 
                         
   /* read the scratchpad */
   flag = ReadTempScratch(rbuf);
   if (flag != 1)
      return 0;

   /* calculate the temperature  */      
   tsht = rbuf[0];
   if (rbuf[1] & 0x01)
      tsht |= -256;
   tmp = (float)(tsht/2);
   cr = rbuf[6];
   cpc = rbuf[7];
   if (rbuf[7] == 0)
      return 0;   
   else
      tmp = tmp - (float)0.25 + (cpc - cr)/cpc;
   
   *current_temp = tmp;
         
   return 1;   
}


/*------------------------------------------------------------------------
 * Read the scratch of a DS18/1920.
 */
short ReadTempScratch(uchar *buf)
{
   short i;
                         
   /* access device */
   if (TMAccess(session_handle,&state_buffer) != 1)
      return 0;
         
   /* send read scratch command */
   TMTouchByte(session_handle,0xBE);
                              
   /* zero the dowcrc */                          
   CRC8 = 0;
   
   /* read scratch */
   for (i = 0; i < 8; i++)   
   {
      buf[i] = (uchar)TMTouchByte(session_handle,(short)0xFF);
      dowcrc(buf[i]);
   }
   
   /* check crc */
   if (dowcrc((uchar)TMTouchByte(session_handle,(short)0xFF)) != 0)
      return 0;
      
   return 1;   
}                   


/*----------------------------------------------------------------------
 * Set the flip flops to the 'ffstate' .
 */                                                          
short SetFlipFlop(short ffstate)
{    
   short printmsg=1,flag,rt=0,cnt=0,done=0;

   do
   {
      /* access and verify it is there */
      if (TMAccess(session_handle,&state_buffer) == 1) 
      {
         /* reset CRC */
         CRC16 = 0;

         /* write status command */
         flag = TMTouchByte(session_handle,0x55);
         docrc16(0x55);

         /* send address */                 
         TMTouchByte(session_handle,0x07);
         docrc16(0x07); 
         TMTouchByte(session_handle,0x00);
         docrc16(0x00);                 
   
         /* write that status */
         TMTouchByte(session_handle,ffstate);
         docrc16(ffstate);                 

         /* crc */
         docrc16(TMTouchByte(session_handle,0xFF));
         docrc16(TMTouchByte(session_handle,0xFF));

         /* check result */
         if (CRC16 == 0xB001)   
            done = TRUE; 
      }
   }
   while (!done && (cnt++ < 10));

   return done;                  
}


/*----------------------------------------------------------------------
 * Read the switch information byte.
 */                                                          
short ReadSwitchInfo(short ClearActivity)
{
   short flag,rt=-1;
   
   /* access and verify it is there */
   flag = TMAccess(session_handle,&state_buffer);
   if (flag == 1) 
   {
      /* reset CRC */
      CRC16 = 0;
   
      /* channel access command */
      flag = TMTouchByte(session_handle,0xF5);
      docrc16(0xF5);
   
      /* control bytes */                 
      if (ClearActivity)
      {
         TMTouchByte(session_handle,0xD5);
         docrc16(0xD5); 
      }
      else
      {
         TMTouchByte(session_handle,0x55);   
         docrc16(0x55);                      
      }
      TMTouchByte(session_handle,0xFF);
      docrc16(0xFF);
   
      /* read the info byte */
      rt = TMTouchByte(session_handle,0xFF);
      docrc16(rt);
   
      /* read a dummy read byte and CRC16 */
      docrc16(TMTouchByte(session_handle,0xFF));
      docrc16(TMTouchByte(session_handle,0xFF));
      docrc16(TMTouchByte(session_handle,0xFF));
   
      /* check result */
      if (CRC16 != 0xB001) 
         rt = -1;
   }
   else
      rt = -1;

   return rt;
}


/*--------------------------------------------------------------------------
 * Update the Dallas Semiconductor One Wire CRC (CRC8) from the global
 * variable CRC8 and the argument.  Return the updated CRC8.
 */
uchar dscrc_table[] = {
        0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
      157,195, 33,127,252,162, 64, 30, 95,  1,227,189, 62, 96,130,220,
       35,125,159,193, 66, 28,254,160,225,191, 93,  3,128,222, 60, 98,
      190,224,  2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
       70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89,  7,
      219,133,103, 57,186,228,  6, 88, 25, 71,165,251,120, 38,196,154,
      101, 59,217,135,  4, 90,184,230,167,249, 27, 69,198,152,122, 36,
      248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91,  5,231,185,
      140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
       17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
      175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
       50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
      202,148,118, 40,171,245, 23, 73,  8, 86,180,234,105, 55,213,139,
       87,  9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
      233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
      116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53};

uchar dowcrc(uchar x)
{
   CRC8 = dscrc_table[CRC8 ^ x];
   return CRC8;
}


/*--------------------------------------------------------------------------
 * Calculate a new CRC16 from the input data shorteger.  Return the current
 * CRC16 and also update the global variable CRC16.
 */
static short oddparity[16] = { 0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0 };

ushort docrc16(ushort data)
{
   data = (data ^ (CRC16 & 0xff)) & 0xff;
   CRC16 >>= 8;

   if (oddparity[data & 0xf] ^ oddparity[data >> 4])
     CRC16 ^= 0xc001;

   data <<= 6;
   CRC16   ^= data;
   data <<= 1;
   CRC16   ^= data;

   return CRC16;
}
 
 
/*--------------------------------------------------------------------------
 *  Prints a message, beeps and exits the program.
 */
void Exit_Prog(char *msg, short errcd)
{
   /* print the message left justified */
   printf("\r%s",msg);
   
   /*  return error code passed */
   exit(errcd);
}


/*--------------------------------------------------------------------------
 *  Conditionally prints a message, 
 */
void output(short print_flag, char *msg)
{
   /* only print following in verbose mode */
   if (print_flag)
   {
      printf("%s",msg);
      if (fp != NULL)
         fprintf(fp,"%s",msg);
   }
}